LEFT_FRONT_WHEEL_SPEED_SENSOR = "lfwss"
RIGHT_FRONT_WHEEL_SPEED_SENSOR = "rfwss"
LEFT_REAR_WHEEL_SPEED_SENSOR = "lrwss"
RIGHT_REAR_WHEEL_SPEED_SENSOR = "rrwss"
MIDDLE_REAR_WHEEL_SPEED_SENSOR = "mrwss"



from os import chmod as _chmod
from os.path import abspath as _abspath


def create_service() -> None:
    _chmod(script := _abspath(__file__)[:-10] + "leads.service.sh", 777)
    with open("/etc/systemd/system/leads.service", "w") as f:
        f.write(f"""
        [Unit]
        Description=LEADS
        After=default.target
        
        [Service]
        ExecStart={script}
        
        [Install]
        WantedBy=default.target
        """)



class Device(object):
    def __init__(self, tag: str, *pins: int | str) -> None:
        self._tag: str = tag
        self._pins: tuple[int | str, ...] = pins

    def tag(self) -> str:
        return self._tag



from argparse import ArgumentParser as _ArgumentParser
from os import mkdir as _mkdir
from os.path import exists as _exists
from sys import exit as _exit, version as _version

from leads_dashboard import get_system_platform as _get_system_platform
from leads_vec.config import load_config as _load_config, DEFAULT_CONFIG as _DEFAULT_CONFIG

if __name__ == '__main__':
    parser = _ArgumentParser(prog="LEADS",
                             description="Lightweight Embedded Assisted Driving System",
                             epilog="GitHub: https://github.com/ProjectNeura/LEADS")
    parser.add_argument("action", choices=("info", "run"), default=None)
    parser.add_argument("-r", "--register", choices=("systemd", "config"), default=None, help="service to register")
    parser.add_argument("-c", "--config", default=None, help="specified configuration file")
    args = parser.parse_args()
    if args.action == "info":
        from leads_vec.__version__ import __version__

        print(f"LEADS Version: {__version__}",
              f"System Platform: {_get_system_platform()}",
              f"Python Version: {_version}",
              sep="\n")
        _exit()
    if args.register == "systemd":
        if _get_system_platform() != "linux":
            _exit("ERROR: Unsupported operating system")
        from ._bootloader import create_service

        if not _exists("/usr/local/leads/config.json"):
            print("Config file not found. Creating \"/usr/local/leads/config.json\"...")
            _mkdir("/usr/local/leads")
            with open("/usr/local/leads/config.json", "w") as f:
                f.write(str(_DEFAULT_CONFIG))
            print("Using \"/usr/local/leads/config.json\"")
        create_service()
        print("Service registered")
    elif args.register == "config":
        if _exists("config.json"):
            r = input("\"config.json\" already exists. Overwrite? (y/N) >>>").lower()
            if r.lower() != "y":
                _exit("ERROR: Aborted")
        with open("config.json", "w") as f:
            f.write(str(_DEFAULT_CONFIG))
        print("Configuration file saved to \"config.json\"")
    try:
        from leads_emulation import SRWRandom as _Controller
    except ImportError:
        raise ImportError("At least one adapter has to be installed")
    config = _load_config(args.config) if args.config else _DEFAULT_CONFIG
    from leads_vec.cli import main

    _exit(main(_Controller("main"), config))



from abc import abstractmethod as _abstractmethod, ABCMeta as _ABCMeta
from json import dumps as _dumps
from time import time as _time
from typing import Self as _Self


class DataContainer(object, metaclass=_ABCMeta):
    def __init__(self) -> None:
        self._time_stamp: int = int(_time() * 1000)

    @_abstractmethod
    def __sub__(self, other: _Self) -> _Self:
        raise NotImplementedError

    def __str__(self) -> str:
        return _dumps(self.to_dict())

    def reset_time_stamp(self) -> None:
        """
        Reset the data's time stamp to now.
        """
        self._time_stamp = _time() * 1000

    def get_time_stamp(self) -> int:
        """
        Get the data's time stamp.
        :return: the time stamp (ms)
        """
        return self._time_stamp

    def to_dict(self) -> dict:
        """
        Convert the data into a dictionary.
        :return: a dictionary that contains all custom attributes of the container
        """
        attributes = dir(self)
        r = {"t": self._time_stamp}
        for n in attributes:
            if n.startswith("_"):
                continue
            v = self.__getattribute__(n)
            if type(v) in (int, float, str):
                r[n] = v
        return r

    def encode(self) -> bytes:
        """
        Encode the data into bytes for network transaction purposes.
        :return: JSON in bytes
        """
        return str(self).encode()


class SRWDataContainer(DataContainer):
    def __init__(self,
                 front_wheel_speed: int | float = 0,
                 rear_wheel_speed: int | float = 0,
                 ) -> None:
        super().__init__()
        self.front_wheel_speed: int | float = front_wheel_speed
        self.rear_wheel_speed: int | float = rear_wheel_speed

    def __sub__(self, other: _Self) -> _Self:
        return SRWDataContainer(
            self.front_wheel_speed - other.front_wheel_speed,
            self.rear_wheel_speed - other.rear_wheel_speed
        )


class DRWDataContainer(DataContainer):
    def __init__(self,
                 front_wheel_speed: int | float = 0,
                 left_rear_wheel_speed: int | float = 0,
                 right_rear_wheel_speed: int | float = 0,
                 ) -> None:
        super().__init__()
        self.front_wheel_speed: int | float = front_wheel_speed
        self.left_rear_wheel_speed: int | float = left_rear_wheel_speed
        self.right_rear_wheel_speed: int | float = right_rear_wheel_speed

    def __sub__(self, other: _Self) -> _Self:
        return DRWDataContainer(
            self.front_wheel_speed - other.front_wheel_speed,
            self.left_rear_wheel_speed - other.left_rear_wheel_speed,
            self.right_rear_wheel_speed - other.right_rear_wheel_speed
        )



from time import time as _time

from leads import DataContainer as _DataContainer
from leads.comm import Client as _Client


class RuntimeData(object):
    start_time: int = int(_time())
    lap_time: list[int] = []
    frame_counter: int = 0
    comm: _Client | None = None

    def comm_notify(self, d: _DataContainer) -> None:
        if self.comm:
            self.comm.send(d.encode())

    def comm_kill(self) -> None:
        if self.comm:
            self.comm.kill()



from json import load as _load, dumps as _dumps
from typing import TextIO as _TextIO, Any as _Any

from leads_dashboard import get_system_platform as _get_system_platform


class Config(object):
    def __init__(self, base: dict[str, _Any]) -> None:
        """
        :param base: base dictionary
        """
        self._d: dict[str, _Any] = base
        self.srw_mode: bool = True
        self.width: int = 720
        self.height: int = 480
        self.fullscreen: bool = False
        self.no_title_bar: bool = False
        self.refresh_rate: int = 30
        self.font_size_small: int = 8
        self.font_size_medium: int = 16
        self.font_size_large: int = 32
        self.font_size_x_large: int = 48
        self.scaling_factor: float = .8 if _get_system_platform() == "linux" else 1
        self.comm_addr: str = "127.0.0.1"
        self.comm_port: int = 16900
        self.data_dir: str = "./data"
        self.refresh()

    def __getitem__(self, name: str) -> _Any | None:
        return self.get(name)

    def __getattr__(self, name: str) -> _Any | None:
        return self.get(name)

    def __setitem__(self, name: str, value: _Any) -> None:
        self._d[name] = value
        self.refresh()

    def __str__(self) -> str:
        return _dumps(self)

    def load(self, d: dict[str, _Any]) -> None:
        """
        Load a new dictionary and refresh.
        :param d: the dictionary to load
        """
        self._d = d
        self.refresh()

    def get(self, name: str, default: _Any | None = None) -> _Any:
        """
        Get the value of a given name from the dictionary.
        :param name: dictionary key
        :param default: default value if the value does not exist
        :return: the value if it exists or else the default value
        """
        if r := self._d.get(name):
            return r
        return default

    def refresh(self) -> None:
        """
        Write the dictionary into the instance attributes.
        """
        for name in dir(self):
            if not name.startswith("_") and (v := self.get(name)):
                self.__setattr__(name, v)


DEFAULT_CONFIG = Config({})


def load_config(file: str | _TextIO) -> Config:
    return Config(_load(open(file)) if isinstance(file, str) else _load(file))
